/*************************************************************************
   format.c: Routines for handling the output format

   Copyright (c) : Alejandro Luque,1,34+954663083

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
************************************************************************
   Parts of this code have been almost literally copied from the source
   code of GNU 'find', written by
   Eric Decker <cire@cisco.com>,
   with enhancements by David MacKenzie <djm@gnu.ai.mit.edu>,
   Jay Plett <jay@silence.princeton.nj.us>,
   Tim Wood <axolotl!tim@toad.com>
   and Dan Bernstein <brnstnd@kramden.acf.nyu.edu>.
************************************************************************/

#include "ising.h"
#include "stdlib.h"
#include "string.h"
#include "ctype.h"
#include "math.h"

extern float kT, h, m, epsilon2, sigma2;
extern int lat_height, lat_width, lat_size;
extern int n_min, N;

/* Creates a new, unlinked segment.
 *
 * Crea un nuevo segmento, sin enlazar.
 */
st_segment *
make_segment (int kind, char *text, int len)
{
  st_segment *new_segment;
  char *fmt;

  new_segment = (st_segment*) xmalloc (sizeof (st_segment) + 2);
  fmt = (char *) xmalloc (len);

  new_segment->kind = kind;
  new_segment->len = len;
  new_segment->text = fmt;

  strncpy (fmt, text, len);

  fmt += len;

  switch (kind) {
  case 't':
  case 'T':
  case 'h':
  case 'H':
  case 'm':
  case 'M':
  case 'e':
  case 's':
    *fmt++ = 'f';
    break;
  case 'i':
  case 'N':
  case 'a':
    *fmt++ = 'd';
    break;
  }

  *fmt = '\0';

  return new_segment;
}

/* Parses a format string to make it understandable by printf and friends.
 *
 * Analiza una cadena para hacerla inteligible a las funciones de tipo
 * printf.
 */
st_segment *
parse_format (char *str)
{
  char *scan, *scan2, *format;
  st_segment *pseg, *curseg, *prevseg;

  pseg = curseg = prevseg = NULL;
  format = str;

  for (scan = str; *scan; scan++) {
    if (*scan == '\\')
	{
	  scan2 = scan + 1;
	  if (*scan2 >= '0' && *scan2 <= '7')
	    {
	      register int n, i;

	      for (i = n = 0; i < 3 && (*scan2 >= '0' && *scan2 <= '7');
		   i++, scan2++)
		n = 8 * n + *scan2 - '0';
	      scan2--;
	      *scan = n;
	    }
	  else
	    {
	      switch (*scan2)
		{
		case 'a':
		  *scan = 7;
		  break;
		case 'b':
		  *scan = '\b';
		  break;
		case 'c':
		  curseg = make_segment (KIND_STOP, format, scan - format);
		  if(prevseg != NULL) prevseg->next = curseg;
		  else pseg = curseg;
		  prevseg = curseg;
		  break;
		case 'f':
		  *scan = '\f';
		  break;
		case 'n':
		  *scan = '\n';
		  break;
		case 'r':
		  *scan = '\r';
		  break;
		case 't':
		  *scan = '\t';
		  break;
		case 'v':
		  *scan = '\v';
		  break;
		case '\\':
		  /* *scan = '\\'; * it already is */
		  break;
		default:
		  scan++;
		  continue;
		}
	    }
	  curseg = make_segment (KIND_PLAIN, format, scan - format + 1);
	  if(prevseg != NULL) prevseg->next = curseg;
	  else pseg = curseg;
	  prevseg = curseg;

	  format = scan2 + 1;	/* Move past the escape. */
	  scan = scan2;		/* Incremented immediately by `for'. */
	}
      else if (*scan == '%')
	{
	  if (scan[1] == '%')
	    {
	      curseg = make_segment (KIND_PLAIN, format, scan - format + 1);
	      if(prevseg != NULL) prevseg->next = curseg;
	      else pseg = curseg;
	      prevseg = curseg;

	      scan++;
	      format = scan + 1;
	      continue;
	    }
	  /* Scan past flags, width and precision, to verify kind. */
	  for (scan2 = scan; *++scan2 && strchr("-+ #", *scan2);)
	    /* Do nothing. */ ;
	  while (isdigit(*scan2))
	    scan2++;
	  if (*scan2 == '.')
	    for (scan2++; isdigit(*scan2); scan2++)
	      /* Do nothing. */ ;
	  if (strchr ("eihtmMHNT", *scan2))
	    {
	      curseg = make_segment (*scan2, format, scan2 - format);
	      if (prevseg != NULL) prevseg->next = curseg;
	      else pseg = curseg;
	      prevseg = curseg;

	      scan = scan2;
	      format = scan + 1;
	    }
	  else
	    {
	      /* An unrecognized % escape.  Print the char after the %. */
	      curseg = make_segment (*scan2, format, scan2 - format);
	      if (prevseg != NULL) prevseg->next = curseg;
	      else pseg = curseg;
	      prevseg = curseg;

	      format = scan + 1;
	      continue;
	    }
	}
    }

  if (scan > format) {
    curseg = make_segment (KIND_PLAIN, format, scan - format);
    if (prevseg != NULL) prevseg->next = curseg;
    else pseg = curseg;
    prevseg = curseg;
  }
  
  return pseg;
}

int seg_fprintf
(FILE *fp, st_segment *segment)
{
  for (; segment; segment = segment->next){
    switch (segment->kind) {
    case KIND_PLAIN:
      fwrite (segment->text, 1, segment->len, fp);
      break;
    case KIND_STOP:
      fwrite (segment->text, 1, segment->len, fp);
      fflush (fp);
      return TRUE;
    case 'H':
    case 'h':
      fprintf (fp, segment->text, h);
      break;
    case 'T':
    case 't':
      fprintf (fp, segment->text, kT);
      break;
    case 'M':
    case 'm':
      fprintf (fp, segment->text, m);
      break;
    case 's':
      fprintf (fp, segment->text, sqrt(sigma2));
      break;
    case 'e':
      fprintf (fp, segment->text, sqrt(epsilon2));
      break;
    case 'i':
      fprintf (fp, segment->text, n_min);
      break;
    case 'N':
      fprintf (fp, segment->text, N);
      break;
    }
  }
  return TRUE;
}

