/*************************************************************************
   lattice.c: Routines for handling the spins lattice.

   Copyright (c) : Alejandro Luque,1,34+954663083

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
************************************************************************/

#include "ising.h"

extern st_spin *lattice;
extern int lat_height, lat_width, lat_size;

/* Applies periodic boundary conditions (pbc) to var _d (0 < _d < _m).
 *
 * Aplica condiciones de contorno periodicas (ccp) a la variable _d. */
#define ccp(_m, _d) (((_m) + (_d)) % (_m))

/* Computes the lattice index of spin located at _x, _y (using pbc).
 *
 * Calcula el indice en la red de la celda _x, _y (aplicando ccp). */
#define lat_index(_x, _y) (ccp (lat_height,_y) * lat_width + ccp (lat_width, _x))

char er_no_mem[] = "Not enought memory";

/* Allocates memory for a a x b lattice of spins.
 *
 * Asigna memoria para almacenar una red de a x b espines. */
st_spin *
lat_alloc (int a, int b)
{
  st_spin *ret;

  ret = (st_spin*) malloc (sizeof (st_spin) * a * b);
  if (ret == NULL)
    fatal (er_no_mem);
  
  return ret;
}

/* Initializes the 'neights' field of a spins lattice.
 *
 * Inicializa el campo neights de una red de espines.
 */
void 
init_neights_4 (void)
{
  int x, y;
  st_spin *ptr;

  for (y = 0, ptr = lattice; y < lat_height; y++)
    for (x = 0; x < lat_width; x++){
      /* We only consideer the four nearest neightbours.
       *
       * Tenemos en cuenta los cuatro vecinos mas proximos */
      ptr->neights[0] = lattice + lat_index (x, y - 1);
      ptr->neights[1] = lattice + lat_index (x, y + 1);
      ptr->neights[2] = lattice + lat_index (x - 1, y);
      (ptr++)->neights[3] = lattice + lat_index (x + 1, y);
    }
}

/* Sets neight_up = 0 for all spins in the lattice.
 *
 * Acera todos los valores de neight_up */
void
init_neight_up (void)
{
  st_spin *ptr;

  for (ptr = lattice; ptr < lattice + lat_size; ptr++)
    ptr->neight_up = 0;
}

/* Initializes the status of the lattice's spins.  p specifies the probability
 * of a given spin to be initially pounting upwards.  p, as all probability
 * measures in the program is normalized to MAX_PROB.  Two other
 * macros are defined concerning probability in ising.h:
 *  * HALF_PROB = A probability of 50%.
 *  * ZERO_PROB = A probability of 0%.
 *
 * Inicializa los estados de la red.  El parametro p especifica la probabilidad
 * de que un espin este inicialmente hacia arriba.  Una probabilidad del 100%
 * corresponde a un valor p = MAX_PROB.  Una red equilibrada corresponde al
 * valor p = HALF_PROB y una probabilidad del 0% corresponde a p = ZERO_PROB.
 * Devuelve el numero de espines inicialmente hacia arriba.
 */
int 
init_status (unsigned int p)
{
  st_spin *ptr;
  int n_up;

  for (ptr = lattice, n_up = 0; ptr < lattice + lat_size; ptr++){
    int i;
    if (mitch_rand () > p) {
      ptr->status = UP;
      n_up ++;
      for (i = 0; i < N_NEIGHTS; i++)
	(ptr->neights[i])->neight_up++;
    } else {
      ptr->status = DOWN;
    }
  }
  return n_up;
}

/* Counts the number of spins currently pointing upwards (magnetization).
 *
 * Cuenta el numero de espines con espin hacia arriba (magnetizacion) */
int
count_n_up (void)
{
  st_spin *ptr;
  int count;

  for (ptr = lattice, count = 0; ptr < lattice + lat_size; ptr++)
    if (ptr->status == UP) count++;

  return count;
}
  
