/*************************************************************************
   loadsave.c: Functions for load/save the lattice.

   Copyright (c) : Alejandro Luque,1,34+954663083

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
************************************************************************
   Some parts of this code are slightly inspired by GNU Octave,
   (c) John W. Eaton <jwe@bevo.che.wisc.edu>, 1996, 1997, 1998
************************************************************************/

#include "ising.h"
#include "stdio.h"
#include "time.h"

/* MATLAB precission codes:
 *
 * Codigos de precision del formato MATLAB:
 */
#define PREC_DOUBLE  0
#define PREC_FLOAT   1
#define PREC_INT     2
#define PREC_SHORT   3
#define PREC_U_SHORT 4
#define PREC_U_CHAR  5

extern st_spin *lattice;
extern int lat_height, lat_width, lat_size;
extern char str_vers[];
extern FILE *fsave, *fload;

char def_var_name[] = "lattice";
char *var_name = def_var_name;
uint32_t var_name_len = sizeof(def_var_name);

char ascii_header[] = 
"# Generated by ising %s %s\n\
# name: ising\n\
# type: matrix\n\
# rows: %d\n\
# columns: %d\n";

char warn_type[] = "wrong type in file header";
char warn_write[] = "unable to write file";
char warn_read[] = "unable to read file";
char warn_name[] = "file contains no name of data. Using default 'lattice'";
char warn_prec[] = "invalid precission specification in file header";
char warn_imag[] = "the matrix for the lattice should be REAL!";
char warn_zero[] = "zero-sized matrix";

/* Prints the lattice to stream fsave.
 *
 * Imprime la matriz en el fichero indicado por el controlador fsave.
 */
int
save_lattice_oct_ascii (void)
{ 
  time_t t;
  st_spin *p;
  int i, j;

  time (&t);

  fseek (fsave, 0, SEEK_SET);
  fprintf (fsave, ascii_header, str_vers, ctime (&t), lat_height, lat_width);

  for (i = 0, p = lattice; i < lat_height; i++) {
    for (j = 0; j < lat_width; j++)
      fprintf (fsave, (j == lat_width - 1) ? "%d" : "%d ", (p++)->status);
    if (i != lat_height - 1) fputc ('\n', fsave);
  }
  fflush (fsave);
  return TRUE;
}

/* Saves the lattice in MATLAB binary format to file stream fsave.
 *
 * Guarda la red en formato binario MATLAB al fichero apuntado por fsave.
 */
int
save_lattice_mat_bin (void)
{
  st_spin *p;

  if (fseek (fsave, 0, SEEK_SET)) {
    warning (warn_write);
    return FALSE;
  }

  if (!put_bin_header ())
    return FALSE;

  for (p = lattice; p < lattice + lat_size; p++){
    /* As there are only two possible states for the spin, we use the smallest
     * data type allowed by MATLAB.  Even thus, we are wasting seven out of
     * eight bits in each byte.
     *
     * Puesto que son solo dos los estados posibles para el espin, utilizaremos
     * el tipo de dato mas reducido permitido por el formato MATLAB.  Aun asi,
     * desperdiciamos siete de cada ocho bits en un byte.
     */
    if (!(xfwrite (&(p -> status), 1, 1, fsave)))
      return FALSE;
  }
  if (fflush (fsave)) {
    warning (warn_write);
    return FALSE;
  }

  return TRUE;
}

/* Puts MATLAB binary header.  Returns TRUE if and only if the function has
 * been succesful.
 *
 * Ecribe una cabezera para un fichero MATLAB binario.  Devuelve TRUE si tiene
 * exito, FALSE en caso contrario.
 */
int
put_bin_header (void)
{
  /* We use the standard ISO C 9X : 7.18 to be sure we use 4 bytes long
   * integers.
   * 
   * Usamos el standard ISO C 9X: 7.18 para asegurarnos de que nuestros
   * enteros son de 4 bytes
   */
  uint32_t var;

  /* mopt (type = unsigned char) */
  var = 50;
  if (!xfwrite (&var, 4, 1, fsave)) 
    return FALSE;

  /* rows */
  var = lat_width;
  if (!xfwrite (&var, 4, 1, fsave))
    return FALSE;

  /* columns */
  var = lat_height;
  if (!xfwrite (&var, 4, 1, fsave))
    return FALSE;

  /* imag */
  var = 0;
  if (!xfwrite (&var, 4, 1, fsave))
    return FALSE;

  /* name_length */
  var = var_name_len;
  if (!xfwrite (&var, 4, 1, fsave))
    return FALSE;

  /* name */
  if (!xfwrite (var_name, 1, var, fsave))
    return FALSE;
  
  return TRUE;
}

/* Loads a matrix in MATLAB binary format and sets it as the lattice.
 *
 * Carga una matriz en formato binario MATLAB y la toma como red.
 */
int
load_lattice_mat_bin (void)
{
  /* See above.
   *
   * Ver arriba.
   */
  uint32_t mopt, rows, columns, imag;
  int type, prec, order, mach;
  spin_t s;
  st_spin *p;

  if (!read_bin_header (&mopt, &rows, &columns, &imag, 
			&var_name_len, &var_name)) 
    return FALSE;

  type = mopt % 10;  /* Full, sparse, etc. */
  mopt /= 10;        /* Eliminate first digit. */
  prec = mopt % 10;  /* double, float, int, etc. */
  mopt /= 10;        /* Eliminate second digit. */
  order = mopt % 10; /* Row or column major ordering. */
  mopt /= 10;        /* Eliminate third digit. */
  mach = mopt % 10;  /* IEEE, VAX, etc. */
  
  if (type != 0) warning (warn_type);
  
  if (imag) {
    warning (warn_imag);
    return FALSE;
  }

  if (!rows || !columns) {
    fatal (warn_zero);
    return FALSE;
  }
      
  if (!order) {
    lat_width = columns;
    lat_height = rows;
  } else {
    lat_width = rows;
    lat_height = columns;
  }

  lattice = lat_alloc (lat_width, lat_height);
  lat_size = lat_width * lat_height;

  init_neights_4 ();
  init_neight_up ();

  for(p = lattice; p < lattice + lat_size; p++) {
    int i;
    s = read_spin (prec);
    if (s == -1) return FALSE;

    p -> status = s;
    if (s == UP) {
      for (i = 0; i < N_NEIGHTS; i++)
	(p->neights[i])->neight_up++;
    }
  }
  return TRUE;
}

int
read_bin_header(uint32_t *mopt, uint32_t *rows, uint32_t *columns,
		     uint32_t *imag, uint32_t *name_length, char **name)
{
  if (!xfread (mopt, 4, 1, fload))
    return FALSE;

  if (!xfread (rows, 4, 1, fload))
    return FALSE;

  if (!xfread (columns, 4, 1, fload))
    return FALSE;

  if (!xfread (imag, 4, 1, fload))
    return FALSE;

  if (!xfread (name_length, 4, 1, fload))
    return FALSE;

  if (!*name_length) {
    warning (warn_name);
    *name_length = sizeof (def_var_name);
    *name = def_var_name;
  } else {
    *name = xmalloc (*name_length);
    if (!xfread (*name, 1, *name_length, fload))
      return FALSE;
  }

  return TRUE;
}

/* Reads a spin from fload with format prec
 *
 * Lee un espin de fload con precision prec
 */
spin_t
read_spin (int prec)
{
  unsigned char dat[8];

  switch (prec) {
  case PREC_DOUBLE:
    if (!xfread (dat, 8, 1, fload)) 
      return -1;
    if (* (double *) dat > 0.5) return UP;
    return DOWN;
  case PREC_FLOAT:
    if (!xfread (dat, 4, 1, fload))
      return -1;
    if (* (float *) dat > 0.5) return UP;
    return DOWN;
  case PREC_INT:
    if (!xfread (dat, 4, 1, fload))
      return -1;
    return (spin_t) (* (int *) dat);
  case PREC_SHORT:
  case PREC_U_SHORT:
    if (!xfread (dat, 2, 1, fload))
      return -1;
    return (spin_t) (* (short *) dat);
  case PREC_U_CHAR:
    if (!xfread (dat, 1, 1, fload))
      return -1;
    return (spin_t) (* (unsigned char *) dat);
  default:
    warning (warn_prec);
    return -1;
  }

  return TRUE;
}

/* I/O functions with error handling.
 *
 * Funciones de E/S con control de errores.
 */
int
xfwrite (void *data, size_t size, size_t count, FILE *stream)
{
  if (fwrite (data, size, count, stream) != count) {
    warning (warn_write);
    return TRUE;
  }
  return FALSE;
}

int
xfread (void *data, size_t size, size_t count, FILE *stream)
{
  if (fread (data, size, count, stream) != count) {
    warning (warn_read);
    return TRUE;
  }
  return FALSE;
}

