/*************************************************************************
   main.c: 'main ()' and related functions of the Ising model simulator

   Copyright (c) : Alejandro Luque,1,34+954663083

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
************************************************************************/

#include "time.h"
#include "getopt.h"
#include "ising.h"

st_spin *lattice;
int lat_height = 128, lat_width = 128, lat_size;
int n_min = DEF_n_min, N = DEF_N, n_sta = DEF_n_sta, 
  seed, reinit_flag = FALSE;

/* iter_total keeps the number of iteration cycles done so far
 * (debugging purposes only).
 *
 * iter_total mantiene el numero total de ciclos de iteraciones
 * (con fines de depuracion).
 */
debug (int iter_total;)

float kT, h, m, sigma2;
/* Be careful, epsilon2 actually contains epsilon^2.
 *
 * Cuidado epsilon2 en realidad almacena el valor de epsilon^2.
 */
float epsilon2 = DEF_epsilon2, *for_var = &h, to, step;

char str_vers[] = VERS;
char def_format[] = "%h %m\n";
char *format_str = def_format;

st_segment *format_seg;

FILE *fout = NULL, *fsave = NULL, *fload = NULL;

char er_no_open[] = "Unable to open file %s";
char er_no_opt[] = "Unknown option %c\n";
char er_inv_for[] = "Invalid 'for' specification";
char er_parse[] = "Parse error in the argument string";
char er_inc_for[] = "Incomplete or mal formed 'for' expression";
char er_N_neg[] = "Can't set N < 0!";
char er_n_min_neg[] = "Can't set negative values for MIN";
char er_n_sta_neg[] = "Can't set negative values for n";
char er_epsilon_neg[] = "Standard deviation should be positive";
char er_dim_neg[] = "Lattice dimensions should be positive";
char er_step_sg[] = "'for' loop would never end";

char warn_reinit[] = "-r, --reinit flag makes no sense without a --for";
char warn_load[] = "Unable to read lattice file: using defaults";
char warn_save[] = "Lattice file not saved";

char msg_vers[] = 
"ising %s\n"
"Copyright (C) 2000 Alejandro Luque Estepa <alques@terra.es>\n"
"You may redistribute copies of this program under the terms of\n"
"the GNU General Public License (see the file GPL)\n";

char msg_help[] = 
"ising %s Bidimensional Ising model simulator\n"
"Alejandro Luque Estepa <alques@terra.es>, 2000\n"
"Cumputes the value of the magnetization for given temperature and external\n"
"magnetic field in a spin lattice as described by the Ising model\n\n"
"  -h, --help                   prints this message and exits.\n"
"  -v, --version                prints the version number and exits.\n"
"  -g, --height=HEIGHT          sets the lattice height.\n"
"  -w, --width=WIDTH            sets the lattice width.\n"
"  -s, --seed=SEED              sets SEED as the seed for the randm number\n"
"                                 generator.  If not present, the sistem time\n"
"                                 is used.\n"
"  -o, --output=FILE            outputs results to file FILE (defaults to\n"
"                                 stdout.\n"
"  -f, --format=STR             sets STR as the output format string. Usual\n"
"                                 C style escape secuences are allowed.\n"
"                                 %%-characters are interpreted as follows:\n"
"                                  %%T, %%t - Prints the system temperature\n"
"                                  %%H, %%h - Prints the magnetic field\n"
"                                  %%M, %%m - Prints the magnetization\n"
"                                  %%i - Prints MIN value (see below)\n"
"                                  %%N - Prints N value (see below)\n"
"                                  %%a - Prints n value (see below)\n"
"                                  %%e - Prints EPS value (see below)\n"
"                                  %%s - Prints the actual standard deviation\n"
"  -L, --load=LOADFILE          loads LOADFILE (in MATLAB binary format) and\n"
"                                 uses it as the initial lttice status.\n"
"  -S, --save=SAVEFILE          saves the lattice in SAVEFILE (in MATLAB\n"
"                                 binary format) every iteration cycle.\n"
"  -T, --temperature=TEMP       sets TEMP as the temperature of the lattice.\n"
"  -H, --field=FIELD            sets FIELD as the value of the magnetic\n"
"                                 field in the lattice.  Both T and h are\n"
"                                 adimensionalized in terms of the elemental\n"
"                                 spin-spin interaction.\n"
"  -N, --iter=N                 sets N as the number of iterations made in\n"
"                                 an interation cycle.  This is roughly\n"
"                                 defined as a number big enought to let the\n"
"                                 lattice have a sensible variation between\n"
"                                 two iteration cycles (by default, N is set\n"
"                                 to 1000000).\n"
"  -i, --min=MIN                sets MIN as the minimum number of iteration\n"
"                                 cycles for given T and h before the\n"
"                                 system is tested to be steady (by default,\n"
"                                 MIN is set to 40).\n"
"  -n, --n-sta=n                the simulator consideers the system to be\n"
"  -e, --epsilon=EPS              steady if the standard cuadratic deviation\n"
"                                 of the magnetization over n iteration\n"
"                                 cycles is less than or equal to EPS\n"
"                                 (default values are n=20, EPS=0.05).\n"
"  -r, --reinit                 reinitializes the lattice when changing\n"
"                                 the values of the temperature or magnetic\n"
"                                 field.  If this option is present, no\n"
"                                 hystheresis phenomena can be observed.\n\n"
"  Different values for temperature or magnetic field can be set up with\n"
"a slightly primitive form of a for/to/step construction, as in\n"
"       ising -H 0.0 --for -T 3.0 --to 2.0 --step -.1\n"
"wich will compute values of the magnetization for a null magnetic field and\n"
"temperature ranging from 3.0 to 2.0 in steps of -0.1.  If the --reinit flag\n"
"is not present, this feature may be used to observe hystheresis phenomena\n\n"
"  Report bugs to alques@terra.es\n";

#ifdef DEBUG
char dbg_n_min[] = "Performing minimum number of iterations...\n";
char dbg_sta[] = "\nWaiting for the lattice to become steady...\n";
char dbg_comp[] = "\nComputing the value of magnetization...\n";
char dbg_iter[] = "%d * %d iterations done\n";
#endif

/* Prints a string and finishes returning -1.
 *
 * Imprime una cadena en stderr y finaliza el programa con el valor -1.
 */
void
fatal (char *str)
{
  fprintf (stderr, "fatal: %s\n", str);
  exit (-1);
}

/* Prints a string to stderr.
 *
 * Imprime una cadena en stderr.
 */
void
warning (char *str)
{
  fprintf (stderr, "warning: %s\n", str);
}

/* xmalloc is defined as recommended in the glibc manual.
 *
 * Definimos una funcion xmalloc tal como se recomienda en el manual
 * de la glibc.
 */
void *
xmalloc (size_t size)
{
  register void *value = malloc (size);
  if (value == 0) fatal ("Virtual memory exhausted");
  return value;
}

int
main (int argc, char **argv)
{
  char *strbuf;
  int c = 0, for_sta = 0;

  seed = time (0);
  debug (setvbuf (stdout, NULL, _IONBF, 0));

  while (c != -1) {
    static struct option long_option[] = {
      {"min", 1, 0, 'i'},
      {"iter", 1, 0, 'N'},
      {"epsilon", 1, 0, 'e'},
      {"n-sta", 1, 0, 'n'},
      {"output", 1, 0, 'o'},
      {"format", 1, 0, 'f'},
      {"seed", 1, 0, 's'},
      {"height", 1, 0, 'g'},
      {"width", 1, 0, 'w'},
      {"temperature", 1, 0, 'T'},
      {"T", 1, 0, 'T'},
      {"field", 1, 0, 'H'},
      {"H", 1, 0, 'H'},
      {"save", 1, 0, 'S'},
      {"load", 1, 0, 'L'},
      {"for", 0, 0, 1000},
      {"to", 1, 0, 2000},
      {"step", 1, 0, 3000},
      {"reinit", 0, 0, 'r'},
      {"version", 0, 0, 'v'},
      {"help", 0, 0, 'h'},
      {0, 0, 0, 0}
    };
    int option_index = 0;

    c = getopt_long (argc, argv, "i:N:e:o:f:s:g:w:a:T:H:S:L:rvh",
		    long_option, &option_index);

    switch (c) {
    case 'i':
      sscanf (optarg, "%d", &n_min);
      if (n_min < 0) fatal (er_n_min_neg);
      break;
    case 'N':
      sscanf (optarg, "%d", &N);
      if (N < 0) fatal (er_N_neg);
      break;
    case 'n':
      sscanf (optarg, "%d", &n_sta);
      if (n_sta < 0) fatal (er_n_sta_neg);
      break;
    case 'e':
      sscanf (optarg, "%f", &epsilon2);
      fatal (er_epsilon_neg);

      epsilon2 *= epsilon2;
      break;
    case 'o':
      if ((fout = fopen (optarg, "a")) == NULL) {
	asprintf (&strbuf, er_no_open, optarg);
	fatal (strbuf); 
      }
      break;
    case 'f':
      format_str = optarg;
      break;
    case 's':
      sscanf (optarg, "%d", &seed);
      break;
    case 'g':
      sscanf (optarg, "%d", &lat_height);
      if (lat_height < 0) fatal (er_dim_neg);
      break;
    case 'w':
      sscanf (optarg, "%d", &lat_width);
      if (lat_width < 0) fatal (er_dim_neg);
      break;
    case 'T':
      if (for_sta > 1) fatal (er_inv_for);
      sscanf (optarg, "%f", &kT);
      if (for_sta == 1) {
	for_var = &kT;
	for_sta ++;
      }
      break;
    case 'H':
      if (for_sta > 1) fatal (er_inv_for);
      sscanf (optarg, "%f", &h);
      if (for_sta == 1) {
	for_var = &h;
	for_sta ++;
      }
      break;
    case 'S':
      if ((fsave = fopen (optarg, "w")) == NULL) {
	asprintf (&strbuf, er_no_open, optarg);
	fatal (strbuf); 
      }
      break;
    case 'L':
      if ((fload = fopen (optarg, "rb")) == NULL) {
	asprintf (&strbuf, er_no_open, optarg);
	fatal (strbuf); 
      }
      break;
    case 1000:
      if (for_sta != 0) fatal (er_parse);
      for_sta = 1;
      break;
    case 2000:
      sscanf (optarg, "%f", &to);
      for_sta++;
      break;
    case 3000:
      sscanf (optarg, "%f", &step);
      for_sta++;
      break;
    case 'r':
      reinit_flag = TRUE;
      break;
    case 'v':
      printf (msg_vers, VERS);
      exit (0);
    case 'h':
      printf (msg_help, str_vers);
      exit (0);
    case '?':
      asprintf (&strbuf, er_no_opt, optopt);
      fatal (strbuf);
      break;
    }
  }

  if (for_sta != 0 && for_sta < 4) 
    fatal (er_inc_for);

  init_global ();

  if (for_sta > 0) {
    if (((to - *for_var) * step) < 0) 
      fatal (er_step_sg);

    for (; (step > 0) ? (*for_var < to) : (*for_var > to); *for_var += step) {
      if (reinit_flag) {
	init_neight_up();
	init_status (HALF_PROB);
      }
      do_it ();
    }
  } else {
    if (reinit_flag) warning (warn_reinit);
    do_it ();
  }

  debug (printf (dbg_iter, iter_total, N));

  return 0;
}

/* Initializes the lattice, the random numbers generator and any other
 * global variable that needs to be initialized.
 *
 * Inicializa la red, el generador de numeros aleatorios y todas las demas
 * variables globales que aun no esten inicializadas.
 */
void
init_global (void)
{  
  mitch_init (seed);

  /* If no file is specified, send output to stdout
   * Si no se especifica un fichero, por defecto escribimos en stdout */
  if (!fout) fout = stdout;

  format_seg = parse_format (format_str);

  debug (iter_total = 0);
  
  init_sta (n_sta);
  if (fload) {
    if (!load_lattice_mat_bin()){
      warning (warn_load);
      fload = NULL;
    }
  }

  if(!fload) {
    lat_size = lat_height * lat_width;
    lattice = lat_alloc (lat_height, lat_width);
    init_neights_4 ();
    init_neight_up ();
    if (!reinit_flag) init_status (HALF_PROB);
  }

  mitch_m_init (seed, lat_size);
}


/* This is the actual main routine.  For T and h given, it computes the
 * magnetization of the lattice, once it suffices the stability conditions,
 * as specified by the user (or default) settings.
 *
 * Esta es la rutina que hace realmente todo el trabajo.  Para unas T y h
 * fijadas, calcula la magnetizacion de la red una vez alcanzadas las
 * condiciones de estabilidad y segun los criterios especificados en las
 * variables globales.
 */
void
do_it (void)
{
  int i;

  debug (printf ("[kT=%.3f h=%.3f]\n", kT, h));

  init_p_table ();
  reset_sta ();

  /* We always perform at least n_min iterations.
   *
   * Hacemos al menos n_min * N iteraciones SIEMPRE */
  debug (printf (dbg_n_min));

  for (i = 0; i < n_min; i++) {
    iterate_N (&m);
    debug (printf ("[%d/%d]\xd", i + 1, n_min));
  }
  debug (iter_total += n_min);
  
  /* Now we wait until the magnetization (n_up) is steady.
   *
   * Ahora esperamos a que se estabilize la magnetizacion (n_up).
   */
   debug (printf (dbg_sta));

   while (!iterate_N (&m)) {
    debug (iter_total += 1);
    debug (printf ("[%f (sigma2 = %f)]\xd", m, sigma2));

   }   

   debug (printf ("\n[m = %.6f]\n\n", m));
   
  /* Finally, we output the result.
   *
   * ..y, por fin, escribimos el resultado. */
  seg_fprintf (fout, format_seg);

  fflush (fout);
}

/* Performs N elemental iterations, computes the magnetization value, 
 * adds it to the history list and sets the average value and returns TRUE
 * if and only if there are elements enought in the history list to compute
 * the average and the squared standard deviation is less than or equal to
 * epsilon.
 *
 * Realiza N iteraciones elementales, calcula el valor de la magnetizacion,
 * lo incluye en la lista historica, fija (si puede) su valor medio y devuelve
 * TRUE si y solo si hay suficientes elementos en la tabla historica para
 * calcular la media y la desviacion cuadratica media al cuadrado es menor
 * que o igual a epsilon.
 */
int
iterate_N (float *m_avg)
{
  int i, n_up;
  float m_i;

  for (i = 0; i < N; i++)
    mc_sim ();

  n_up = count_n_up ();
  m_i = 2 * (float) n_up / lat_size - 1.0;

  /* If the user has asked us to do so, we save the current lattice status.
   *
   * Si el usuario nos lo ha pedido, grabamos el estado de la red.
   */
  if (fsave) 
    if (!save_lattice_mat_bin ())
      warning (warn_save);

  if (!add_sta (m_i, m_avg, &sigma2))
    return FALSE;

  if (sigma2 > epsilon2)
    return FALSE;
 
  return TRUE;
}
      











