/*************************************************************************
   mcsim.c: Montecarlo simulator and statistical routines.

   Copyright (c) : Alejandro Luque,1,34+954663083

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
************************************************************************/

#include "ising.h"
#include "stdio.h"
#include "math.h"
#include "errno.h"

extern st_spin *lattice;
extern int lat_size;
extern float kT, h;

float *hist;
int hist_size, hist_i;

/* When the hamiltonian increment for an eventual inversion of a spin
 * is dH > 0, the probability of the spin being inverted is
 * p = exp(-dH / kT).  Now, for given T and h, and for every possible
 * status of the spin, p depends only on the number of next neightbours
 * that point upward so there are only N_NEIGHTS + 1 possibilities
 * for each status of the spin.
 * Therefore we will compute the possible values of p (normalized to
 * MAX_PROB) only once for every T and h, and record them on p_table.
 * If for a configuration of n upward-pointing neightbours and for a status
 * s of the spin, dH < 0, then we set p_table[s][n] = -1.
 *
 * Cuando el incremento del hamiltoniano para la posible inversion de un 
 * espin es dH > 0, la probabilidad de que el espin se invierta es
 * p = exp(-dH / kT).  Ahora bien, para unas T y h dadas, y para cada 
 * estado del espin que consideremos, esta probabilidad
 * depende unicamente del numero de espines vecinos orientados paralelamente
 * al espin considerado, en total, N_NEIGHTS + 1 posibilidades por cada estado
 * de espin (arriba o abajo).
 * Asi pues, para no repetir el calculo de la exponencial, almacenamos 
 * la probabilidad p (normalizada a MAX_PROB) en la tabla p_table.
 * Por otro lado, si para una configuracion de n vecinos hacia arriba y para
 * una orientacion del espin s, dH < 0 hacemos p_table[s][n] = -1.
 */
int p_table[2][N_NEIGHTS + 1];

void
init_p_table (void)
{
  int i, s;
  float dH;
  for (i = 0; i < N_NEIGHTS + 1; i++) {
    for (s = 0; s < 2; s++) {
      dH = delta_ham (s, i);
      if (dH < 0) p_table[s][i] = -1;
      else p_table[s][i] = (int) (MAX_PROB * exp (-dH/kT));
    }
  }
}
 
/* Chooses a randomly selected spin and applies the Montecarlo method to
 * wonder if the orientation should be inverted.  Returns TRUE or FALSE
 * depending on wether the spin has changed or not.
 *   
 * Escoje una celda cualquiera y aplica el metodo de Montecarlo para decidir
 * si cambiar o no el estado de su espin.  Devuelve TRUE o FALSE
 * dependiendo de si ha cambiado o no algun espin de la red.
 */
int
mc_sim (void)
{
  st_spin *s;
  int p;

  s = lattice + mitch_m_rand ();
  p = p_table[(unsigned int) s->status][s->neight_up]; 

  if (p < 0) {
      toggle_spin (s);
      return TRUE;
  }
  if (mitch_rand () < p){
      toggle_spin (s);
      return TRUE;
  }
  return FALSE;
}

/* Returns the hamiltonian increment that would happen if we invert
 * a spin in status s and with n upward-pointing neightbours.
 *
 * Devuelve el incremento que se produciria en el hamiltoniano del
 * sistema si cambiaramos el estado de un espin en estado s con n vecinos
 * hacia arriba.
 */
float
delta_ham (spin_t s, int n)
{
    return (s ? (2 * ((2 * n - N_NEIGHTS) + h)) : 
	    (2 * ((N_NEIGHTS - 2 * n) - h)));
}

/* Inverts the spin pointed by s.
 *
 * Invierte la orientacion del espin apuntado por s.
 */
void
toggle_spin(st_spin *s)
{
    int i;
    if (s->status == DOWN) {
	s->status =  (spin_t) UP;
	for (i = 0; i < N_NEIGHTS; i++)
	    (s->neights[i])->neight_up++;
    } else {
	s->status = (spin_t) DOWN;
	for (i = 0; i < N_NEIGHTS; i++)
	    (s->neights[i])->neight_up--;
    }
}
	
/* Initializes the statistical table.
 *
 * Inicializa la tabla de valores de la magnetizacion.
 */
void
init_sta (int size)
{
  hist = (float *) xmalloc (sizeof (float) * size);
  hist_size = size;
  reset_sta ();
}

/* Resets the history list.
 *
 * Reinicializa la lista historica.
 */
void
reset_sta (void)
{
  float *p;
  for (p = hist; p < hist + hist_size; p++)
    *p = NAN;
  hist_i = 0;
}

/* Adds a value to the history table and returns the average and deviation
 * of existent values.  Returns TRUE if succesful, FALSE if unsuccesful
 * (because the table is not still complete)
 *
 * Incluye un dato mas en la tabla historica y calcula el valor medio y
 * la desviacion media cuadratica de los valores existentes.
 */
int
add_sta (float new_val, float *avg, float *sigma2)
{
  float Sx, Sx2;
  float *p;

  *(hist + hist_i++) = new_val;
  hist_i %= hist_size;

  for (p = hist, Sx = 0, Sx2 = 0; p < hist + hist_size; p++) {
    if (isnanf (*p)) return FALSE;
    Sx += *p;
    Sx2 += (*p) * (*p);
  }

  *avg = Sx / hist_size;
  *sigma2 = Sx2 / hist_size - (*avg) * (*avg);
  
  return TRUE;
}






